#include <msp/core/algorithm.h>
#include <msp/core/hash.h>
#include <msp/core/raii.h>
#include <msp/strings/lexicalcast.h>
#include "finalize.h"
#include "glsl_error.h"
#include "reflect.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

StructOrganizer::StructOrganizer():
	offset(-1)
{ }

void StructOrganizer::visit(StructDeclaration &strct)
{
	SetForScope<int> set_offset(offset, 0);
	TraversingVisitor::visit(strct);
}

void StructOrganizer::visit(VariableDeclaration &var)
{
	if(offset>=0)
	{
		int *layout_offset = 0;
		bool has_matrix_order = false;
		if(var.layout)
		{
			vector<Layout::Qualifier> &qualifiers = var.layout->qualifiers;
			for(vector<Layout::Qualifier>::iterator i=qualifiers.begin(); i!=qualifiers.end(); ++i)
			{
				if(i->name=="offset" && i->has_value)
				{
					layout_offset = &i->value;
					if(i->value>=offset)
						offset = i->value;
				}
				else if(i->name=="column_major" || i->name=="row_major")
					has_matrix_order = true;
			}
		}

		MemoryRequirementsCalculator::Result mem_reqs = MemoryRequirementsCalculator().apply(var);
		offset += mem_reqs.alignment-1;
		offset -= offset%mem_reqs.alignment;

		if(layout_offset)
			*layout_offset = offset;
		else
		{
			if(!var.layout)
				var.layout = new Layout;

			Layout::Qualifier qual;
			qual.name = "offset";
			qual.has_value = true;
			qual.value = offset;
			var.layout->qualifiers.push_back(qual);
		}

		if(!has_matrix_order)
		{
			const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(var.type_declaration);
			while(basic && basic->kind==BasicTypeDeclaration::ARRAY)
				basic = dynamic_cast<const BasicTypeDeclaration *>(basic->base_type);
			if(basic && basic->kind==BasicTypeDeclaration::MATRIX)
			{
				Layout::Qualifier qual;
				qual.name = "column_major";
				var.layout->qualifiers.push_back(qual);
			}
		}

		offset += mem_reqs.size;
	}
}


void LocationAllocator::apply(Module &module, const Features &features)
{
	for(list<Stage>::iterator i=module.stages.begin(); i!=module.stages.end(); ++i)
		apply(*i);
	allocate_locations("uniform");

	for(vector<InterfaceBlock *>::const_iterator i=unbound_blocks.begin(); i!=unbound_blocks.end(); ++i)
		bind_uniform((*i)->layout, (*i)->block_name, features.uniform_binding_range);
	for(vector<VariableDeclaration *>::const_iterator i=unbound_textures.begin(); i!=unbound_textures.end(); ++i)
		bind_uniform((*i)->layout, (*i)->name, features.texture_binding_range);
}

void LocationAllocator::apply(Stage &stage)
{
	swap(used_locations["in"], used_locations["out"]);
	used_locations["out"].clear();

	stage.content.visit(*this);

	allocate_locations("in");
	allocate_locations("out");
}

void LocationAllocator::allocate_locations(const string &iface)
{
	vector<VariableDeclaration *>::iterator write = unplaced_variables.begin();
	unsigned next = 0;
	for(vector<VariableDeclaration *>::const_iterator i=unplaced_variables.begin(); i!=unplaced_variables.end(); ++i)
	{
		if((*i)->interface!=iface)
		{
			if(write!=i)
				*write = *i;
			++write;
			continue;
		}

		if((*i)->interface=="uniform")
		{
			map<string, Uniform>::const_iterator j = uniforms.find((*i)->name);
			if(j!=uniforms.end() && j->second.location>=0)
			{
				add_layout_value((*i)->layout, "location", j->second.location);
				continue;
			}
		}

		set<unsigned> &used = used_locations[(*i)->interface];

		unsigned size = LocationCounter().apply(**i);
		while(1)
		{
			int blocking = -1;
			for(unsigned j=0; j<size; ++j)
				if(used.count(next+j))
					blocking = next+j;
			if(blocking<0)
				break;
			next = blocking+1;
		}

		add_layout_value((*i)->layout, "location", next);
		if((*i)->interface=="uniform")
			uniforms[(*i)->name].location = next;

		for(unsigned j=0; j<size; ++j)
			used.insert(next+j);
		next += size;
	}

	unplaced_variables.erase(write, unplaced_variables.end());
}

void LocationAllocator::bind_uniform(RefPtr<Layout> &layout, const string &name, unsigned range)
{
	map<string, Uniform>::const_iterator i = uniforms.find(name);
	if(i!=uniforms.end() && i->second.bind_point>=0)
		add_layout_value(layout, "binding", i->second.bind_point);
	else
	{
		set<unsigned> &used = used_bindings[0];

		unsigned bind_point = fold32(hash64(name))%range;
		while(used.count(bind_point))
			bind_point = (bind_point+1)%range;

		add_layout_value(layout, "binding", bind_point);
		uniforms[name].bind_point = bind_point;
		used.insert(bind_point);
	}
}

void LocationAllocator::add_layout_value(RefPtr<Layout> &layout, const string &name, unsigned value)
{
	if(!layout)
		layout = new Layout;

	Layout::Qualifier qual;
	qual.name = name;
	qual.has_value = true;
	qual.value = value;
	layout->qualifiers.push_back(qual);
}

void LocationAllocator::visit(VariableDeclaration &var)
{
	if(!var.name.compare(0, 3, "gl_"))
		return;

	if(!var.interface.empty())
	{
		int location = (var.layout ? get_layout_value(*var.layout, "location") : -1);

		if(location<0 && var.linked_declaration && var.linked_declaration->layout)
		{
			location = get_layout_value(*var.linked_declaration->layout, "location");
			if(location>=0)
				add_layout_value(var.layout, "location", location);
		}

		if(location>=0)
		{
			unsigned size = LocationCounter().apply(var);
			for(unsigned i=0; i<size; ++i)
				used_locations[var.interface].insert(location+i);
			if(var.interface=="uniform")
				uniforms[var.name].location = location;
		}
		else
			unplaced_variables.push_back(&var);
	}

	if(var.interface=="uniform")
	{
		const TypeDeclaration *type = var.type_declaration;
		while(const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(type))
			type = basic->base_type;
		if(dynamic_cast<const ImageTypeDeclaration *>(type))
		{
			int bind_point = (var.layout ? get_layout_value(*var.layout, "binding") : -1);
			if(bind_point>=0)
			{
				used_bindings[0].insert(bind_point);
				uniforms[var.name].bind_point = bind_point;
			}
			else
				unbound_textures.push_back(&var);
		}
	}
}

void LocationAllocator::visit(InterfaceBlock &iface)
{
	if(!iface.instance_name.compare(0, 3, "gl_"))
		return;

	if(iface.interface=="uniform")
	{
		int bind_point = (iface.layout ? get_layout_value(*iface.layout, "binding") : -1);

		if(bind_point>=0)
		{
			used_bindings[0].insert(bind_point);
			uniforms[iface.block_name].bind_point = bind_point;
		}
		else
			unbound_blocks.push_back(&iface);
	}
}


PrecisionConverter::PrecisionConverter():
	stage(0)
{ }

void PrecisionConverter::apply(Stage &s)
{
	stage = &s;
	s.content.visit(*this);
	NodeRemover().apply(s, nodes_to_remove);
}

void PrecisionConverter::visit(Block &block)
{
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(&block==&stage->content)
			insert_point = i;
		(*i)->visit(*this);
	}
}

void PrecisionConverter::visit(Precision &prec)
{
	if(stage->required_features.gl_api==OPENGL_ES2)
		have_default.insert(prec.type);
	else
		nodes_to_remove.insert(&prec);
}

void PrecisionConverter::visit(VariableDeclaration &var)
{
	if(stage->required_features.gl_api!=OPENGL_ES2)
	{
		var.precision.clear();
		return;
	}

	const char *default_prec = (stage->type==Stage::FRAGMENT ? "mediump" : "highp");
	const TypeDeclaration *type = var.type_declaration;
	while(type)
	{
		if(dynamic_cast<const ImageTypeDeclaration *>(type))
		{
			default_prec = "lowp";
			break;
		}
		else if(const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(type))
		{
			if(basic->kind==BasicTypeDeclaration::INT || basic->kind==BasicTypeDeclaration::FLOAT)
				break;
			type = basic->base_type;
		}
		else
			return;
	}
	if(!type)
		return;

	if(!have_default.count(type->name))
	{
		Precision *prec = new Precision;
		prec->precision = default_prec;
		prec->type = type->name;
		stage->content.body.insert(insert_point, prec);

		have_default.insert(type->name);
	}
}


LegacyConverter::LegacyConverter():
	frag_out(0)
{ }

void LegacyConverter::apply(Stage &s, const Features &feat)
{
	stage = &s;
	features = feat;
	if(supports_stage(s.type))
	{
		s.content.visit(*this);
		NodeRemover().apply(s, nodes_to_remove);

		if(!stage->required_features.glsl_version)
			stage->required_features.glsl_version = Version(1, (stage->required_features.gl_api==OPENGL_ES2 ? 0 : 10));
	}
	else
		unsupported(format("Stage %s is not supported", Stage::get_stage_name(s.type)));
}

void LegacyConverter::unsupported(const string &reason)
{
	Diagnostic diagnostic;
	diagnostic.severity = Diagnostic::ERR;
	diagnostic.source = GENERATED_SOURCE;
	diagnostic.line = 0;
	diagnostic.message = reason;
	stage->diagnostics.push_back(diagnostic);
}

void LegacyConverter::visit(Block &block)
{
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(&block==&stage->content)
			uniform_insert_point = i;
		(*i)->visit(*this);
	}
}

bool LegacyConverter::check_version(const Version &feature_version) const
{
	if(features.glsl_version<feature_version)
		return false;
	else if(stage->required_features.glsl_version<feature_version)
		stage->required_features.glsl_version = feature_version;

	return true;
}

bool LegacyConverter::check_extension(bool Features::*extension) const
{
	if(!(features.*extension))
		return false;

	stage->required_features.*extension = true;

	return true;
}

bool LegacyConverter::supports_stage(Stage::Type st) const
{
	if(st==Stage::GEOMETRY)
	{
		if(features.gl_api==OPENGL_ES2)
			return check_version(Version(3, 20));
		else
			return check_version(Version(1, 50));
	}
	else
		return true;
}

bool LegacyConverter::supports_unified_interface_syntax() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else
		return check_version(Version(1, 30));
}

void LegacyConverter::visit(VariableReference &var)
{
	if(var.declaration==frag_out && !supports_unified_interface_syntax())
	{
		var.name = "gl_FragColor";
		var.declaration = 0;
	}
}

void LegacyConverter::visit(Assignment &assign)
{
	TraversingVisitor::visit(assign);
	if(assign.target.declaration==frag_out && !supports_unified_interface_syntax())
		assign.target.declaration = 0;
}

bool LegacyConverter::supports_unified_sampling_functions() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else
		return check_version(Version(1, 30));
}

void LegacyConverter::visit(FunctionCall &call)
{
	if(call.declaration && call.declaration->source==BUILTIN_SOURCE)
	{
		if(!call.name.compare(0, 7, "texture") && call.arguments.size()>=1)
		{
			const ImageTypeDeclaration *arg_image = dynamic_cast<const ImageTypeDeclaration *>(call.arguments.front()->type);
			if(arg_image && !supports_unified_sampling_functions())
			{
				string suffix = call.name.substr(7);
				call.name = (arg_image->shadow ? "shadow" : "texture");

				switch(arg_image->dimensions)
				{
				case ImageTypeDeclaration::ONE: call.name += "1D"; break;
				case ImageTypeDeclaration::TWO: call.name += "2D"; break;
				case ImageTypeDeclaration::THREE: call.name += "3D"; break;
				case ImageTypeDeclaration::CUBE: call.name += "Cube"; break;
				}

				if(arg_image->array)
				{
					/* Array textures and the unified sampling function name were
					both introduced in GLSL 1.30. */
					if(arg_image->dimensions==ImageTypeDeclaration::ONE || arg_image->dimensions==ImageTypeDeclaration::TWO)
						check_extension(&Features::ext_texture_array);
					call.name += "Array";
				}

				call.name += suffix;
			}
		}
	}

	TraversingVisitor::visit(call);
}

bool LegacyConverter::supports_interface_layouts() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else if(check_version(Version(3, 30)))
		return true;
	else if(check_version(Version(1, 30)))
		return check_extension(&Features::arb_explicit_attrib_location);
	else
		return false;
}

bool LegacyConverter::supports_stage_interface_layouts() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 10));
	else if(check_version(Version(4, 10)))
		return true;
	else
		return check_extension(&Features::arb_separate_shader_objects);
}

bool LegacyConverter::supports_centroid_sampling() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 0));
	else if(check_version(Version(1, 20)))
		return true;
	else
		return check_extension(&Features::ext_gpu_shader4);
}

bool LegacyConverter::supports_sample_sampling() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 20));
	else if(check_version(Version(4, 0)))
		return true;
	else
		return check_extension(&Features::arb_gpu_shader5);
}

bool LegacyConverter::supports_uniform_location() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 10));
	else if(check_version(Version(4, 30)))
		return true;
	else
		return check_extension(&Features::arb_explicit_uniform_location);
}

bool LegacyConverter::supports_binding() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 10));
	else
		return check_version(Version(4, 20));
}

void LegacyConverter::visit(VariableDeclaration &var)
{
	if(var.layout)
	{
		for(vector<Layout::Qualifier>::const_iterator i=var.layout->qualifiers.begin(); i!=var.layout->qualifiers.end(); )
		{
			if(i->name=="location")
			{
				bool supported = true;
				bool external = false;
				if(var.interface=="in")
				{
					external = (stage->type==Stage::VERTEX);
					supported = (external ? supports_interface_layouts() : supports_stage_interface_layouts());
				}
				else if(var.interface=="out")
				{
					external = (stage->type==Stage::FRAGMENT);
					supported = (external ? supports_interface_layouts() : supports_stage_interface_layouts());
					if(external && !supported && !check_extension(&Features::ext_gpu_shader4))
					{
						external = false;
						if(i->value!=0)
							unsupported("EXT_gpu_shader4 required for multiple fragment shader outputs");
					}
				}
				else if(var.interface=="uniform")
					supported = supports_uniform_location();

				if(!supported)
				{
					if(external)
						stage->locations[var.name] = i->value;
					i = var.layout->qualifiers.erase(i);
				}
				else
					++i;
			}
			else if(i->name=="binding" && !supports_binding())
			{
				const TypeDeclaration *type = var.type_declaration;
				while(const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(type))
					type = basic->base_type;
				if(dynamic_cast<const ImageTypeDeclaration *>(type))
					stage->texture_bindings[var.name] = i->value;

				i = var.layout->qualifiers.erase(i);
			}
			else
				++i;
		}

		if(var.layout->qualifiers.empty())
			var.layout = 0;
	}

	if(var.sampling=="centroid")
	{
		if(!supports_centroid_sampling())
			var.sampling = string();
	}
	else if(var.sampling=="sample")
	{
		if(!supports_sample_sampling())
			var.sampling = string();
	}

	if((var.interface=="in" || var.interface=="out") && !supports_unified_interface_syntax())
	{
		if(stage->type==Stage::FRAGMENT && var.interface=="out")
		{
			frag_out = &var;
			nodes_to_remove.insert(&var);
		}
	}

	TraversingVisitor::visit(var);
}

bool LegacyConverter::supports_interface_blocks(const string &iface) const
{
	if(features.gl_api==OPENGL_ES2)
	{
		if(iface=="uniform")
			return check_version(Version(3, 0));
		else
			return check_version(Version(3, 20));
	}
	else if(check_version(Version(1, 50)))
		return true;
	else if(iface=="uniform")
		return check_extension(&Features::arb_uniform_buffer_object);
	else
		return false;
}

bool LegacyConverter::supports_interface_block_location() const
{
	if(features.gl_api==OPENGL_ES2)
		return check_version(Version(3, 20));
	else if(check_version(Version(4, 40)))
		return true;
	else
		return check_extension(&Features::arb_enhanced_layouts);
}

void LegacyConverter::visit(InterfaceBlock &iface)
{
	if(iface.layout)
	{
		for(vector<Layout::Qualifier>::const_iterator i=iface.layout->qualifiers.begin(); i!=iface.layout->qualifiers.end(); )
		{
			if(i->name=="location" && !supports_interface_block_location())
				i = iface.layout->qualifiers.erase(i);
			else if(i->name=="binding" && !supports_binding())
			{
				stage->uniform_block_bindings[iface.block_name] = i->value;
				i = iface.layout->qualifiers.erase(i);
			}
			else
				++i;
		}

		if(iface.layout->qualifiers.empty())
			iface.layout = 0;
	}

	if(!supports_interface_blocks(iface.interface) && iface.type_declaration)
	{
		if(!iface.instance_name.empty())
			unsupported("ARB_uniform_buffer_object required for interface block instances");
		else if(iface.struct_declaration)
		{
			stage->content.body.splice(uniform_insert_point, iface.struct_declaration->members.body);
			nodes_to_remove.insert(&iface);
			nodes_to_remove.insert(iface.struct_declaration);
		}
		else
			/* If the interface block is an array, it should have an instance
			name too, so this should never be reached */
			throw logic_error("Unexpected interface block configuration");
	}
}

} // namespace SL
} // namespace GL
} // namespace Msp
