#include <algorithm>
#include <cstring>
#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "reflect.h"
#include "validate.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

Validator::Validator():
	stage(0),
	last_provoker(0)
{ }

void Validator::diagnose(Node &node, Node &provoking_node, Diagnostic::Severity severity, const string &message)
{
	Diagnostic diag;
	diag.severity = severity;
	diag.source = node.source;
	diag.line = node.line;
	diag.provoking_source = provoking_node.source;
	diag.provoking_line = provoking_node.line;
	diag.message = message;
	stage->diagnostics.push_back(diag);

	last_provoker = &provoking_node;
}

void Validator::add_info(Node &node, const string &message)
{
	if(!last_provoker)
		throw logic_error("Tried to add info without a previous provoker");
	diagnose(node, *last_provoker, Diagnostic::INFO, message);
}


DeclarationValidator::DeclarationValidator():
	scope(GLOBAL),
	iface_layout(0),
	iface_block(0),
	variable(0)
{ }

const char *DeclarationValidator::describe_variable(ScopeType scope)
{
	switch(scope)
	{
	case GLOBAL: return "global variable";
	case STRUCT: return "struct member";
	case INTERFACE_BLOCK: return "interface block member";
	case FUNCTION_PARAM: return "function parameter";
	case FUNCTION: return "local variable";
	default: return "variable";
	}
}

void DeclarationValidator::visit(Layout &layout)
{
	for(vector<Layout::Qualifier>::const_iterator i=layout.qualifiers.begin(); i!=layout.qualifiers.end(); ++i)
	{
		bool allowed = false;
		string err_descr;
		bool value = true;
		if(i->name=="location")
			allowed = (variable && scope==GLOBAL);
		else if(i->name=="binding" || i->name=="set")
		{
			if(i->name=="set")
			{
				error(layout, "Layout qualifier 'set' not allowed when targeting OpenGL");
				continue;
			}

			if(variable)
			{
				TypeDeclaration *type = variable->type_declaration;
				while(BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(type))
					type = basic->base_type;
				bool uniform = (variable->interface=="uniform");
				allowed = (scope==GLOBAL && uniform && dynamic_cast<ImageTypeDeclaration *>(type));
				err_descr = (uniform ? "variable of non-opaque type" : "non-uniform variable");
			}
			else if(iface_block)
			{
				allowed = (iface_block->interface=="uniform");
				err_descr = "non-uniform interface block";
			}
		}
		else if(i->name=="constant_id")
		{
			allowed = (variable && scope==GLOBAL);
			if(allowed)
			{
				if(!variable->constant)
				{
					allowed = false;
					err_descr = "non-constant variable";
				}
				else
				{
					BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(variable->type_declaration);
					if(!basic || basic->kind<BasicTypeDeclaration::BOOL || basic->kind>BasicTypeDeclaration::INT)
					{
						allowed = false;
						err_descr = format("variable of type '%s'",
							(variable->type_declaration ? variable->type_declaration->name : variable->type));
					}
				}
			}
		}
		else if(i->name=="offset")
			allowed = (variable && scope==INTERFACE_BLOCK && iface_block->interface=="uniform");
		else if(i->name=="align")
			allowed = (scope==INTERFACE_BLOCK && iface_block->interface=="uniform");
		else if(i->name=="points")
		{
			allowed = (stage->type==Stage::GEOMETRY && iface_layout && (iface_layout->interface=="in" || iface_layout->interface=="out"));
			value = false;
		}
		else if(i->name=="lines" || i->name=="lines_adjacency" || i->name=="triangles" || i->name=="triangles_adjacency")
		{
			allowed = (stage->type==Stage::GEOMETRY && iface_layout && iface_layout->interface=="in");
			value = false;
		}
		else if(i->name=="line_strip" || i->name=="triangle_strip")
		{
			allowed = (stage->type==Stage::GEOMETRY && iface_layout && iface_layout->interface=="out");
			value = false;
		}
		else if(i->name=="invocations")
			allowed = (stage->type==Stage::GEOMETRY && iface_layout && iface_layout->interface=="in");
		else if(i->name=="max_vertices")
			allowed = (stage->type==Stage::GEOMETRY && iface_layout && iface_layout->interface=="out");
		else if(i->name=="std140" || i->name=="std430")
		{
			allowed = (iface_block && !variable && iface_block->interface=="uniform");
			value = false;
		}
		else if(i->name=="column_major" || i->name=="row_major")
		{
			allowed = (variable && scope==INTERFACE_BLOCK);
			if(allowed)
			{
				BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(variable->type_declaration);
				while(basic && basic->kind==BasicTypeDeclaration::ARRAY)
					basic = dynamic_cast<BasicTypeDeclaration *>(basic->base_type);
				allowed = (basic && basic->kind==BasicTypeDeclaration::MATRIX);
				err_descr = "non-matrix variable";
			}
		}

		if(!allowed)
		{
			if(err_descr.empty())
			{
				if(variable)
					err_descr = describe_variable(scope);
				else if(iface_block)
					err_descr = "interface block";
				else if(iface_layout)
					err_descr = format("interface '%s'", iface_layout->interface);
				else
					err_descr = "unknown declaration";
			}
			error(layout, format("Layout qualifier '%s' not allowed on %s", i->name, err_descr));
		}
		else if(value && !i->has_value)
			error(layout, format("Layout qualifier '%s' requires a value", i->name));
		else if(!value && i->has_value)
			error(layout, format("Layout qualifier '%s' does not allow a value", i->name));
	}
}

void DeclarationValidator::visit(InterfaceLayout &layout)
{
	SetForScope<InterfaceLayout *> set_layout(iface_layout, &layout);
	TraversingVisitor::visit(layout);
}

void DeclarationValidator::visit(BasicTypeDeclaration &type)
{
	BasicTypeDeclaration *basic_base = dynamic_cast<BasicTypeDeclaration *>(type.base_type);
	BasicTypeDeclaration::Kind base_kind = (basic_base ? basic_base->kind : BasicTypeDeclaration::VOID);

	if(type.kind==BasicTypeDeclaration::VECTOR && (base_kind<BasicTypeDeclaration::BOOL || base_kind>BasicTypeDeclaration::FLOAT))
		error(type, format("Invalid base type '%s' for vector type '%s'", type.base, type.name));
	else if(type.kind==BasicTypeDeclaration::MATRIX && base_kind!=BasicTypeDeclaration::VECTOR)
		error(type, format("Invalid base type '%s' for matrix type '%s'", type.base, type.name));
	else if(type.kind==BasicTypeDeclaration::ARRAY && basic_base && base_kind==BasicTypeDeclaration::VOID)
		error(type, format("Invalid base type '%s' for array type '%s'", type.base, type.name));
}

void DeclarationValidator::visit(ImageTypeDeclaration &type)
{
	BasicTypeDeclaration::Kind base_kind = BasicTypeDeclaration::VOID;
	if(BasicTypeDeclaration *basic_base = dynamic_cast<BasicTypeDeclaration *>(type.base_type))
		base_kind = basic_base->kind;
	if(base_kind!=BasicTypeDeclaration::INT && base_kind!=BasicTypeDeclaration::FLOAT)
		error(type, format("Invalid base type '%s' for image type '%s'", type.base, type.name));
}

void DeclarationValidator::visit(StructDeclaration &strct)
{
	SetForScope<ScopeType> set_scope(scope, (scope!=INTERFACE_BLOCK ? STRUCT : scope));
	TraversingVisitor::visit(strct);
}

void DeclarationValidator::visit(VariableDeclaration &var)
{
	SetForScope<VariableDeclaration *> set_var(variable, &var);

	const char *descr = describe_variable(scope);

	if(var.layout)
	{
		if(scope!=GLOBAL && scope!=INTERFACE_BLOCK)
			error(var, format("Layout qualifier not allowed on %s", descr));
		else
			var.layout->visit(*this);
	}

	if(var.constant)
	{
		if(scope==STRUCT || scope==INTERFACE_BLOCK)
			error(var, format("Constant qualifier not allowed on %s", descr));
		if(!var.init_expression)
			error(var, "Constant variable must have an initializer");
	}

	if(!var.interpolation.empty() || !var.sampling.empty())
	{
		if(var.interface!="in" && stage->type==Stage::VERTEX)
			error(var, "Interpolation qualifier not allowed on vertex input");
		else if(var.interface!="out" && stage->type==Stage::FRAGMENT)
			error(var, "Interpolation qualifier not allowed on fragment output");
		else if((var.interface!="in" && var.interface!="out") || (scope==FUNCTION_PARAM || scope==FUNCTION))
			error(var, "Interpolation qualifier not allowed on non-interpolated variable");
	}

	if(!var.interface.empty())
	{
		if(iface_block && var.interface!=iface_block->interface)
			error(var, format("Mismatched interface qualifier '%s' inside '%s' block", var.interface, iface_block->interface));
		else if(scope==STRUCT || scope==FUNCTION)
			error(var, format("Interface qualifier not allowed on %s", descr));
	}

	TypeDeclaration *type = var.type_declaration;
	while(BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(type))
		type = basic->base_type;
	if(dynamic_cast<ImageTypeDeclaration *>(type))
	{
		if(scope!=GLOBAL && scope!=FUNCTION_PARAM)
			error(var, format("Type '%s' not allowed on %s", type->name, descr));
		else if(scope==GLOBAL && var.interface!="uniform")
			error(var, format("Type '%s' only allowed with uniform interface", type->name));
	}

	if(var.init_expression)
	{
		if(scope==GLOBAL && !var.constant)
			error(var, format("Initializer not allowed on non-constant %s", descr));
		else if(scope!=GLOBAL && scope!=FUNCTION)
			error(var, format("Initializer not allowed on %s", descr));
		else
			var.init_expression->visit(*this);
	}
}

void DeclarationValidator::visit(InterfaceBlock &iface)
{
	SetForScope<ScopeType> set_scope(scope, INTERFACE_BLOCK);
	SetForScope<InterfaceBlock *> set_iface(iface_block, &iface);

	if(stage->type==Stage::VERTEX && iface.interface=="in")
		error(iface, "Interface block not allowed on vertex shader input");
	else if(stage->type==Stage::FRAGMENT && iface.interface=="out")
		error(iface, "Interface block not allowed on fragment shader output");

	TraversingVisitor::visit(iface);
	if(iface.struct_declaration)
		iface.struct_declaration->visit(*this);
}

void DeclarationValidator::visit(FunctionDeclaration &func)
{
	SetForScope<ScopeType> set_scope(scope, FUNCTION_PARAM);
	for(NodeArray<VariableDeclaration>::const_iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
		(*i)->visit(*this);
	scope = FUNCTION;
	func.body.visit(*this);
}


IdentifierValidator::IdentifierValidator():
	anonymous_block(false)
{ }

void IdentifierValidator::multiple_definition(const string &name, Statement &statement, Statement &previous)
{
	error(statement, format("Multiple definition of %s", name));
	add_info(previous, "Previous definition is here");
}

Statement *IdentifierValidator::find_definition(const string &name)
{
	BlockDeclarationMap *decls = &declarations[current_block];
	BlockDeclarationMap::const_iterator i = decls->find(name);
	if(i==decls->end() && anonymous_block)
	{
		decls = &declarations[current_block->parent];
		i = decls->find(name);
	}
	return (i!=decls->end() ? i->second : 0);
}

void IdentifierValidator::check_definition(const string &name, Statement &statement)
{
	if(Statement *previous = find_definition(name))
		multiple_definition(format("'%s'", name), statement, *previous);
	else
		record_definition(name, statement);
}

void IdentifierValidator::record_definition(const string &name, Statement &statement)
{
	declarations[current_block].insert(make_pair(name, &statement));
	if(anonymous_block)
		declarations[current_block->parent].insert(make_pair(name, &statement));
}

void IdentifierValidator::visit(TypeDeclaration &type)
{
	check_definition(type.name, type);
}

void IdentifierValidator::visit(StructDeclaration &strct)
{
	check_definition(strct.name, strct);
	TraversingVisitor::visit(strct);
}

void IdentifierValidator::visit(VariableDeclaration &var)
{
	check_definition(var.name, var);
	TraversingVisitor::visit(var);
}

void IdentifierValidator::visit(InterfaceBlock &iface)
{
	string key = iface.interface+iface.block_name;
	map<string, InterfaceBlock *>::const_iterator i = interface_blocks.find(key);
	if(i!=interface_blocks.end())
		multiple_definition(format("interface block '%s %s'", iface.interface, iface.block_name), iface, *i->second);
	else
		interface_blocks.insert(make_pair(key, &iface));

	if(Statement *previous = find_definition(iface.block_name))
	{
		if(!dynamic_cast<InterfaceBlock *>(previous))
			multiple_definition(format("'%s'", iface.block_name), iface, *previous);
	}
	else
		record_definition(iface.block_name, iface);

	if(!iface.instance_name.empty())
		check_definition(iface.instance_name, iface);

	if(iface.instance_name.empty() && iface.struct_declaration)
	{
		// Inject anonymous interface block members into the global scope
		const map<string, VariableDeclaration *> &iface_vars = iface.struct_declaration->members.variables;
		for(map<string, VariableDeclaration *>::const_iterator j=iface_vars.begin(); j!=iface_vars.end(); ++j)
			check_definition(j->first, *j->second);
	}
}

void IdentifierValidator::visit(FunctionDeclaration &func)
{
	string key = func.name+func.signature;
	map<string, FunctionDeclaration *>::const_iterator i = overloaded_functions.find(key);
	if(i==overloaded_functions.end())
		overloaded_functions.insert(make_pair(key, &func));
	else if(func.return_type_declaration && i->second->return_type_declaration!=func.return_type_declaration)
	{
		error(func, format("Conflicting return type '%s' for function '%s'", func.return_type_declaration->name, func.name));
		if(i->second->return_type_declaration)
			add_info(*i->second, format("Previously declared as returning '%s'", i->second->return_type_declaration->name));
	}

	if(Statement *previous = find_definition(func.name))
	{
		if(!dynamic_cast<FunctionDeclaration *>(previous))
			multiple_definition(format("'%s'", func.name), func, *previous);
	}
	else
		record_definition(func.name, func);

	if(func.definition==&func)
		check_definition(func.name+func.signature, func);

	TraversingVisitor::visit(func);
}


void ReferenceValidator::visit(BasicTypeDeclaration &type)
{
	if(!type.base.empty() && !type.base_type)
		error(type, format("Use of undeclared type '%s'", type.base));
}

void ReferenceValidator::visit(ImageTypeDeclaration &type)
{
	if(!type.base.empty() && !type.base_type)
		error(type, format("Use of undeclared type '%s'", type.base));
}

void ReferenceValidator::visit(VariableReference &var)
{
	if(!var.declaration)
		error(var, format("Use of undeclared variable '%s'", var.name));
	else if(stage->type!=Stage::VERTEX && var.declaration->interface=="in" && var.name.compare(0, 3, "gl_") && !var.declaration->linked_declaration)
		error(var, format("Use of unlinked input variable '%s'", var.name));
}

void ReferenceValidator::visit(MemberAccess &memacc)
{
	if(memacc.left->type && !memacc.declaration)
		error(memacc, format("Use of undeclared member '%s'", memacc.member));
	TraversingVisitor::visit(memacc);
}

void ReferenceValidator::visit(InterfaceBlockReference &iface)
{
	/* An interface block reference without a declaration should be impossible
	since references are generated based on existing declarations. */
	if(!iface.declaration)
		error(iface, format("Use of undeclared interface block '%s'", iface.name));
	else if(stage->type!=Stage::VERTEX && iface.declaration->interface=="in" && !iface.declaration->linked_block)
		error(iface, format("Use of unlinked input block '%s'", iface.name));
}

void ReferenceValidator::visit(FunctionCall &call)
{
	if((!call.constructor && !call.declaration) || (call.constructor && !call.type))
	{
		bool have_declaration = call.constructor;
		if(!call.constructor)
		{
			map<string, FunctionDeclaration *>::iterator i = stage->functions.lower_bound(call.name);
			have_declaration = (i!=stage->functions.end() && i->second->name==call.name);
		}

		if(have_declaration)
		{
			bool valid_types = true;
			string signature;
			for(NodeArray<Expression>::const_iterator j=call.arguments.begin(); (valid_types && j!=call.arguments.end()); ++j)
			{
				if((*j)->type)
					append(signature, ", ", (*j)->type->name);
				else
					valid_types = false;
			}

			if(valid_types)
				error(call, format("No matching %s found for '%s(%s)'", (call.constructor ? "constructor" : "overload"), call.name, signature));
		}
		else
			error(call, format("Call to undeclared function '%s'", call.name));
	}
	TraversingVisitor::visit(call);
}

void ReferenceValidator::visit(VariableDeclaration &var)
{
	if(!var.type_declaration)
		error(var, format("Use of undeclared type '%s'", var.type));
	TraversingVisitor::visit(var);
}

void ReferenceValidator::visit(InterfaceBlock &iface)
{
	if(!iface.struct_declaration)
		error(iface, format("Interface block '%s %s' lacks a struct declaration", iface.interface, iface.block_name));
	TraversingVisitor::visit(iface);
}

void ReferenceValidator::visit(FunctionDeclaration &func)
{
	if(!func.return_type_declaration)
		error(func, format("Use of undeclared type '%s'", func.return_type));
	TraversingVisitor::visit(func);
}


ExpressionValidator::ExpressionValidator():
	current_function(0)
{ }

void ExpressionValidator::visit(Swizzle &swizzle)
{
	unsigned size = 0;
	if(BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(swizzle.left->type))
	{
		if(basic->kind==BasicTypeDeclaration::INT || basic->kind==BasicTypeDeclaration::FLOAT)
			size = 1;
		else if(basic->kind==BasicTypeDeclaration::VECTOR)
			size = basic->size;
	}

	if(size)
	{
		static const char component_names[] = { 'x', 'y', 'z', 'w', 'r', 'g', 'b', 'a', 's', 't', 'p', 'q' };
		int flavour = -1;
		for(unsigned i=0; i<swizzle.count; ++i)
		{
			unsigned component_flavour = (find(component_names, component_names+12, swizzle.component_group[i])-component_names)/4;
			if(flavour==-1)
				flavour = component_flavour;
			else if(flavour>=0 && component_flavour!=static_cast<unsigned>(flavour))
			{
				error(swizzle, format("Flavour of swizzle component '%c' is inconsistent with '%c'",
					swizzle.component_group[i], swizzle.component_group[0]));
				flavour = -2;
			}

			if(swizzle.components[i]>=size)
				error(swizzle, format("Access to component '%c' which is not present in '%s'",
					swizzle.component_group[i], swizzle.left->type->name));
		}
	}
	else if(swizzle.left->type)
		error(swizzle, format("Swizzle applied to '%s' which is neither a scalar nor a vector", swizzle.left->type->name));

	TraversingVisitor::visit(swizzle);
}

void ExpressionValidator::visit(UnaryExpression &unary)
{
	if(unary.expression->type)
	{
		if(!unary.type)
			error(unary, format("No matching operator '%s' found for '%s'", unary.oper->token, unary.expression->type->name));
		else if((unary.oper->token[1]=='+' || unary.oper->token[1]=='-') && !unary.expression->lvalue)
			error(unary, format("Operand of '%s' is not an lvalue", unary.oper->token));
	}
	TraversingVisitor::visit(unary);
}

void ExpressionValidator::visit(BinaryExpression &binary)
{
	if(!binary.type && binary.left->type && binary.right->type)
	{
		if(binary.oper->token[0]=='[')
			error(binary, format("Can't index element of '%s' with '%s'",
				binary.left->type->name, binary.right->type->name));
		else
			error(binary, format("No matching operator '%s' found for '%s' and '%s'",
				binary.oper->token, binary.left->type->name, binary.right->type->name));
	}
	TraversingVisitor::visit(binary);
}

void ExpressionValidator::visit(Assignment &assign)
{
	if(assign.left->type)
	{
		if(!assign.left->lvalue)
			error(assign, "Target of assignment is not an lvalue");
		if(assign.right->type)
		{
			if(assign.oper->token[0]!='=')
			{
				if(!assign.type)
					error(assign, format("No matching operator '%s' found for '%s' and '%s'",
						string(assign.oper->token, strlen(assign.oper->token)-1), assign.left->type->name, assign.right->type->name));
			}
			else if(assign.left->type!=assign.right->type)
				error(assign, format("Assignment to variable of type '%s' from expression of incompatible type '%s'",
					assign.left->type->name, assign.right->type->name));
		}
	}
	TraversingVisitor::visit(assign);
}

void ExpressionValidator::visit(TernaryExpression &ternary)
{
	if(ternary.condition->type)
	{
		BasicTypeDeclaration *basic_cond = dynamic_cast<BasicTypeDeclaration *>(ternary.condition->type);
		if(!basic_cond || basic_cond->kind!=BasicTypeDeclaration::BOOL)
			error(ternary, "Ternary operator condition is not a boolean");
		else if(!ternary.type && ternary.true_expr->type && ternary.false_expr->type)
			error(ternary, format("Ternary operator has incompatible types '%s' and '%s'",
				ternary.true_expr->type->name, ternary.false_expr->type->name));
	}
	TraversingVisitor::visit(ternary);
}

void ExpressionValidator::visit(VariableDeclaration &var)
{
	if(var.init_expression && var.init_expression->type && var.type_declaration && var.init_expression->type!=var.type_declaration)
		error(var, format("Initializing a variable of type '%s' with an expression of incompatible type '%s'",
			var.type_declaration->name, var.init_expression->type->name));
	TraversingVisitor::visit(var);
}

void ExpressionValidator::visit(FunctionDeclaration &func)
{
	SetForScope<FunctionDeclaration *> set_func(current_function, &func);
	TraversingVisitor::visit(func);
}

void ExpressionValidator::visit(Return &ret)
{
	if(current_function && current_function->return_type_declaration)
	{
		TypeDeclaration *return_type = current_function->return_type_declaration;
		BasicTypeDeclaration *basic_return = dynamic_cast<BasicTypeDeclaration *>(return_type);
		if(ret.expression)
		{
			if(ret.expression->type && ret.expression->type!=return_type)
				error(ret, format("Return expression type '%s' is incompatible with declared return type '%s'",
					ret.expression->type->name, return_type->name));
		}
		else if(!basic_return || basic_return->kind!=BasicTypeDeclaration::VOID)
			error(ret, "Return statement without an expression in a function not returning 'void'");
	}

	TraversingVisitor::visit(ret);
}


int StageInterfaceValidator::get_location(const Layout &layout)
{
	for(vector<Layout::Qualifier>::const_iterator i=layout.qualifiers.begin(); i!=layout.qualifiers.end(); ++i)
		if(i->name=="location")
			return i->value;
	return -1;
}

void StageInterfaceValidator::visit(VariableDeclaration &var)
{
	int location = (var.layout ? get_location(*var.layout) : -1);
	if(var.interface=="in" && var.linked_declaration)
	{
		const Layout *linked_layout = var.linked_declaration->layout.get();
		int linked_location = (linked_layout ? get_location(*linked_layout) : -1);
		if(linked_location!=location)
		{
			error(var, format("Mismatched location %d for 'in %s'", location, var.name));
			add_info(*var.linked_declaration, format("Linked to 'out %s' with location %d",
				var.linked_declaration->name, linked_location));
		}
		if(var.type_declaration && var.linked_declaration->type_declaration)
		{
			const TypeDeclaration *type = var.type_declaration;
			if(stage->type==Stage::GEOMETRY)
			{
				if(const BasicTypeDeclaration *basic = dynamic_cast<const BasicTypeDeclaration *>(type))
					if(basic->kind==BasicTypeDeclaration::ARRAY && basic->base_type)
						type = basic->base_type;
			}
			if(!is_same_type(*type, *var.linked_declaration->type_declaration))
			{
				error(var, format("Mismatched type '%s' for 'in %s'", type->name, var.name));
				add_info(*var.linked_declaration, format("Linked to 'out %s' with type '%s'",
					var.linked_declaration->name, var.linked_declaration->type_declaration->name));
			}
		}
	}

	if(location>=0 && !var.interface.empty())
	{
		map<unsigned, VariableDeclaration *> &used = used_locations[var.interface];

		unsigned loc_count = LocationCounter().apply(var);
		for(unsigned i=0; i<loc_count; ++i)
		{
			map<unsigned, VariableDeclaration *>::const_iterator j = used.find(location+i);
			if(j!=used.end())
			{
				error(var, format("Overlapping location %d for '%s %s'", location+i, var.interface, var.name));
				add_info(*j->second, format("Previously used here for '%s %s'", j->second->interface, j->second->name));
			}
			else
				used[location+i] = &var;
		}
	}
}


void GlobalInterfaceValidator::apply(Module &module)
{
	for(list<Stage>::iterator i=module.stages.begin(); i!=module.stages.end(); ++i)
	{
		stage = &*i;
		i->content.visit(*this);
	}
}

void GlobalInterfaceValidator::check_uniform(const Uniform &uni)
{
	map<std::string, const Uniform *>::const_iterator i = used_names.find(uni.name);
	if(i!=used_names.end())
	{
		if(uni.location>=0 && i->second->location>=0 && i->second->location!=uni.location)
		{
			error(*uni.node, format("Mismatched location %d for uniform '%s'", uni.location, uni.name));
			add_info(*i->second->node, format("Previously declared here with location %d", i->second->location));
		}
		if(uni.bind_point>=0 && i->second->bind_point>=0 && i->second->bind_point!=uni.bind_point)
		{
			error(*uni.node, format("Mismatched binding %d for uniform '%s'", uni.bind_point, uni.name));
			add_info(*i->second->node, format("Previously declared here with binding %d", i->second->bind_point));
		}
		if(uni.type && i->second->type && !is_same_type(*uni.type, *i->second->type))
		{
			string type_name = (dynamic_cast<const StructDeclaration *>(uni.type) ?
				"structure" : format("type '%s'", uni.type->name));
			error(*uni.node, format("Mismatched %s for uniform '%s'", type_name, uni.name));

			string message = "Previously declared here";
			if(!dynamic_cast<const StructDeclaration *>(i->second->type))
				message += format(" with type '%s'", i->second->type->name);
			add_info(*i->second->node, message);
		}
	}
	else
		used_names.insert(make_pair(uni.name, &uni));

	if(uni.location>=0)
	{
		map<unsigned, const Uniform *>::const_iterator j = used_locations.find(uni.location);
		if(j!=used_locations.end())
		{
			if(j->second->name!=uni.name)
			{
				error(*uni.node, format("Overlapping location %d for '%s'", uni.location, uni.name));
				add_info(*j->second->node, format("Previously used here for '%s'", j->second->name));
			}
		}
		else
		{
			for(unsigned k=0; k<uni.loc_count; ++k)
				used_locations.insert(make_pair(uni.location+k, &uni));
		}
	}

	if(uni.bind_point>=0)
	{
		map<unsigned, const Uniform *> &used = used_bindings[uni.desc_set];
		map<unsigned, const Uniform *>::const_iterator j = used.find(uni.bind_point);
		if(j!=used.end())
		{
			if(j->second->name!=uni.name)
			{
				error(*uni.node, format("Overlapping binding %d for '%s'", uni.bind_point, uni.name));
				add_info(*j->second->node, format("Previously used here for '%s'", j->second->name));
			}
		}
		else
			used.insert(make_pair(uni.bind_point, &uni));
	}
}

void GlobalInterfaceValidator::visit(VariableDeclaration &var)
{
	if(var.interface=="uniform")
	{
		Uniform uni;
		uni.node = &var;
		uni.type = var.type_declaration;
		uni.name = var.name;
		if(var.layout)
		{
			uni.location = get_layout_value(*var.layout, "location");
			uni.loc_count = LocationCounter().apply(var);
			uni.desc_set = get_layout_value(*var.layout, "set", 0);
			uni.bind_point = get_layout_value(*var.layout, "binding");
		}

		uniforms.push_back(uni);
		check_uniform(uniforms.back());
	}
}

void GlobalInterfaceValidator::visit(InterfaceBlock &iface)
{
	if(iface.interface=="uniform")
	{
		Uniform uni;
		uni.node = &iface;
		uni.type = iface.struct_declaration;
		uni.name = iface.block_name;
		if(iface.layout)
		{
			uni.desc_set = get_layout_value(*iface.layout, "set", 0);
			uni.bind_point = get_layout_value(*iface.layout, "binding");
		}

		uniforms.push_back(uni);
		check_uniform(uniforms.back());
	}
}

} // namespace SL
} // namespace GL
} // namespace Msp
